/*
 * libvoodoo.c, (c) 2002 John Clemens
 * Distributed under the terms and conditions of the GNU GPL.
 *
 * Library for manipulating aspects of 3Dfx Voodoo series cards.
 */


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>

#include <voodoo.h>
#include <pci.h>

#define TDFX_REFFREQ  14.31818 /* We'll use Mhz */
#define TDFX_PLLCTRL0 0x10 /* byte offset 0x40 */
#define TDFX_PLLCTRL1 0x11 /* byte offset 0x44 */
#define TDFX_PLLCTRL2 0x12 /* byte offset 0x48 */

/* interal functions */
static int get_pll(struct voodoo_info *v, int reg);
static void set_pll(struct voodoo_info *v, int reg, int pll);
static int freq_to_pll(int freq, int minm, int maxm);
static int pll_to_freq(int pll);
static int map_mmio(struct voodoo_info *v);
static void unmap_mmio(struct voodoo_info *v, int fd);

static int pll_to_freq(int pll)
{
	int freq;
        int N = ((pll & 0x0000FF00) >> 8);
        int M = ((pll & 0x000000FC) >> 2);
        int K = (pll & 0x00000003);

        freq = (int)((TDFX_REFFREQ * (N+2))/((M+2)*(1 << K)));
        return freq;
}

/* This code is slightly modified from the one in XFree86's tdfx_driver.c */
static int freq_to_pll(int freq, int minm, int maxm)
{
          int m, n, k, best_m, best_n, best_k, f_cur, best_error;

          freq = freq*1000; /* convert to khz herre for more significant bits */

          best_error=freq;
          best_n=best_m=best_k=0;

          if (minm < 0) minm = 1;
          if (maxm < 0) maxm = 64;

          for (n=1; n<255; n++) {
                  f_cur=(TDFX_REFFREQ*1000)*(n+2);
                  if (f_cur < freq) {
                          f_cur=f_cur/3;
                          if (freq-f_cur<best_error) {
                                  best_error=freq-f_cur;
                                  best_n=n;
                                  best_m=1;
                                  best_k=0;
                                  continue;
                          }
                  }
                  for (m=minm; m<maxm; m++) {
                          for (k=0; k<4; k++) {
                                  f_cur=(TDFX_REFFREQ*1000)*
                                                          (n+2)/(m+2)/(1<<k);
                                  if (abs(f_cur-freq) < best_error) {
                                          best_error=abs(f_cur-freq);
                                          best_n=n;
                                          best_m=m;
                                          best_k=k;
                                  }
                          }
                  }
          }
          n=best_n;
          m=best_m;
          k=best_k;
          
          return (n<<8)|(m<<2)|k;
}

static int get_pll(struct voodoo_info *v, int reg)
{
        return ((unsigned int)(*(v->mmio+reg)));
}

static void set_pll(struct voodoo_info *v, int reg, int pll)
{
        (*(v->mmio+reg)) = pll;
}

static int map_mmio(struct voodoo_info *v)
{
        int length = 0x0007FFFF;
        int fd;

        if ((fd = open("/dev/mem", O_RDWR)) == -1) {
                perror("Can't open /dev/mem for writing, are you root?");
                return(-EPERM);
        }  

        if ((v->mmio = mmap(0,length,PROT_READ|PROT_WRITE,MAP_SHARED,
             fd,v->mmio_addr)) == (void *)-1) {
                     perror("Can't mmap /dev/mem for rw, are you root?");
                     close(fd);
                     return(-EPERM);
        }
             
        return (fd);
}

static void unmap_mmio(struct voodoo_info *v, int fd)
{
        int length = 0x0007FFFF;
        munmap(v->mmio, length);
        close(fd);
        return;
}

/* public functions */

int voodoo_get_gfx_speed(struct voodoo_info *v)
{
        int fd;
        int pll;

        if ((v->pci_dev_id == TDFX_BANSHEE) ||
                        (v->pci_dev_id == TDFX_VOODOO3)) {
                fd = map_mmio(v);
                if (fd < 0) {
                        return(-EPERM);
                }

                pll=get_pll(v, TDFX_PLLCTRL1);
                unmap_mmio(v, fd);

                return (pll_to_freq(pll));
        } else {
                printf("Don't know how to get GFX speed for your card :(\n");
        }

        return(-1);
}

int voodoo_set_gfx_speed(struct voodoo_info *v, int speed)
{
        int fd;
        int pll;

        if ((v->pci_dev_id == TDFX_BANSHEE) ||
                        (v->pci_dev_id == TDFX_VOODOO3)) {
                fd = map_mmio(v);
                if (fd < 0) {
                        return(-EPERM);
                }

                if (v->pci_dev_id == TDFX_BANSHEE)
                        pll=freq_to_pll(speed, 24, 24);
                else
                        pll=freq_to_pll(speed, 0, 0);

                set_pll(v, TDFX_PLLCTRL1, pll);

                unmap_mmio(v, fd);

                return (pll_to_freq(pll));
        } else {
                printf("Don't know how to get GFX speed for your card :(\n");
        }

        return(-1);
}

int voodoo_get_mem_speed(struct voodoo_info *v)
{
        int fd;
        int pll;


        if ((v->pci_dev_id == TDFX_BANSHEE) || (v->pci_dev_id == TDFX_VOODOO3)) {
                fd = map_mmio(v);
                if (fd < 0) {
                        return(-EPERM);
                }

                if (v->pci_dev_id == TDFX_VOODOO3)
                        pll=get_pll(v, TDFX_PLLCTRL1);
                else
                        pll=get_pll(v, TDFX_PLLCTRL2);

                unmap_mmio(v, fd);

                return (pll_to_freq(pll));
        } else {
                printf("Don't know how to get GFX speed for your card :(\n");
        }

        return(-1);
}

int voodoo_set_mem_speed(struct voodoo_info *v, int speed)
{
        int fd;
        int pll;


        if ((v->pci_dev_id == TDFX_BANSHEE) || (v->pci_dev_id == TDFX_VOODOO3)) {
                fd = map_mmio(v);
                if (fd < 0) {
                        return(-EPERM);
                }

                if (v->pci_dev_id == TDFX_BANSHEE) {
                        pll = freq_to_pll(speed, 24, 24);
                        set_pll(v, TDFX_PLLCTRL2, pll);
                } else {
                        pll = freq_to_pll(speed, 0, 0);
                        set_pll(v, TDFX_PLLCTRL1, pll);
                }

                unmap_mmio(v, fd);

                return (pll_to_freq(pll));
        } else {
                printf("Don't know how to get Mem speed for your card :(\n");
        }

        return(-1);
}


struct voodoo_info * voodoo_probe(void)
{
        struct voodoo_info *v, *cur, *t1;
        struct pci_access *pacc;
        struct pci_dev *dev;
        int count = 0;

        v = (struct voodoo_info *)NULL;

        pacc = pci_alloc();
        pci_init(pacc);
        pci_scan_bus(pacc);

        for (dev=pacc->devices; dev; dev=dev->next) {
                pci_fill_info(dev, PCI_FILL_IDENT | PCI_FILL_BASES);

                if (dev->vendor_id != PCI_VENDOR_TDFX)
                        continue;

                /* We've found a 3dfx card, now allocate a new info struct */
                cur = (struct voodoo_info *)malloc(sizeof(struct voodoo_info));
                if (cur == NULL) {
                        perror("Can't allocate memory for info structure\n");
                        return ((struct voodoo_info *)-ENOMEM);
                }

                /* populate it */
                cur->next = (struct voodoo_info *)NULL;
                cur->pci_dev_id = dev->device_id;
                cur->mmio_addr = dev->base_addr[0];
                cur->memory_size = dev->size[1];
                cur->mmio = (void *)NULL;
                cur->default_gfx_speed = 0;
                cur->default_memory_speed = 0;
                cur->gfx_speed = 0;
                cur->memory_speed = 0;

                /* hook it into the current list */
                if (!v)
                        v = cur;
                else {
                        t1 = v;
                        while(t1->next != NULL) t1 = t1->next;
                        t1->next = cur;
                }
                count++;
        }

        pci_cleanup(pacc);

	if (count == 0) return ((struct voodoo_info *)NULL);
	
        t1 = v;

        while (t1) {
                t1->default_gfx_speed = voodoo_get_gfx_speed(t1);
                t1->default_memory_speed = voodoo_get_mem_speed(t1);
                t1->gfx_speed = t1->default_gfx_speed;
                t1->memory_speed = t1->default_memory_speed;
                switch(t1->pci_dev_id) {
                         case TDFX_VOODOO:
                                  strncpy(t1->name, "3dfx Voodoo", 50);
                                  break;
                         case TDFX_VOODOO2:
                                  strncpy(t1->name, "3dfx Voodoo2", 50);
                                  break;
                         case TDFX_BANSHEE:
                                  strncpy(t1->name, "3dfx Voodoo Banshee", 50);
                                  break;
                         case TDFX_VELOCITY:
                                  strncpy(t1->name, "3dfx Voodoo Velocity", 50);
                                  break;
                         case TDFX_VOODOO3:
                                  strncpy(t1->name, "3dfx Voodoo3", 50);
                                  break;
                         case TDFX_VOODOO5:
                                  strncpy(t1->name, "3dfx Voodoo4/5", 50);
                                  break;
                }
		t1 = t1->next;
        }        

        return (v);
}

int voodoo_cleanup(struct voodoo_info *v)
{
        struct voodoo_info *t1;

        while (v) {
                t1 = v->next;
                free(v);
                v = t1;
        }
        return(0);
}
                

        
        
